// concat.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#define PIPE_TIMEOUT 5000
#define BUFSIZE 4096
#define ENVNAME "CONCATPID"
#define PIPENAMESIZE 128

int startPipeServer();
int connectToPipeServer();
int dualpump(HANDLE input, HANDLE output);
DWORD WINAPI threadpump(LPVOID lpvParam);
int singlepump(HANDLE input, HANDLE output);
void watchpeer(HANDLE hProcess);
DWORD WINAPI threadwatchpeer(LPVOID lpvParam);

int _tmain(int argc, _TCHAR* argv[])
{
	size_t requiredSize;
	_wgetenv_s( &requiredSize, NULL, 0, TEXT(ENVNAME));

	if (requiredSize == 0) return startPipeServer();

	if (argc > 1) SetConsoleTitle(argv[1]);

	return connectToPipeServer();
}


int startPipeServer()
{
	TCHAR sPid[20];
	swprintf_s(sPid, 20, L"%i", GetCurrentProcessId());

	int r = _wputenv_s(TEXT(ENVNAME), sPid);
	if( r != 0 )
	{
		fwprintf( stderr, L"Failed to set " TEXT(ENVNAME) L"\n" );
		return 1;
	}

	TCHAR sPipeM[PIPENAMESIZE];
	TCHAR sPipeS[PIPENAMESIZE];

	swprintf_s(sPipeM, PIPENAMESIZE, L"\\\\.\\pipe\\concat-%s-M", sPid);
	swprintf_s(sPipeS, PIPENAMESIZE, L"\\\\.\\pipe\\concat-%s-S", sPid);

	HANDLE hPipeM = CreateNamedPipe( 
		sPipeM,                   // pipe name 
		PIPE_ACCESS_INBOUND,      // read/write access 
		PIPE_TYPE_BYTE |          // byte-type pipe 
		PIPE_READMODE_BYTE |      // byte read mode 
		PIPE_WAIT,                // blocking mode 
		1,                        // 1 instance 
		0,                        // output buffer size 
		0,                        // input buffer size 
		PIPE_TIMEOUT,             // client time-out 
		NULL);                    // default security attributes
	if (hPipeM == INVALID_HANDLE_VALUE) 
	{
		fwprintf( stderr, L"CreateNamedPipe failed with %d.\n", GetLastError()); 
		return 2;
	}

	HANDLE hPipeS = CreateNamedPipe( 
		sPipeS,                   // pipe name 
		PIPE_ACCESS_OUTBOUND,     // read/write access 
		PIPE_TYPE_BYTE |          // byte-type pipe 
		PIPE_READMODE_BYTE |      // byte read mode 
		PIPE_WAIT,                // blocking mode 
		1,                        // 1 instance 
		0,                        // output buffer size 
		0,                        // input buffer size 
		PIPE_TIMEOUT,             // client time-out 
		NULL);                    // default security attributes
	if (hPipeS == INVALID_HANDLE_VALUE) 
	{
		fwprintf( stderr, L"CreateNamedPipe failed with %d.\n", GetLastError()); 
		return 3;
	}

	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	memset(&si, 0, sizeof(si)); 
	memset(&pi, 0, sizeof(pi)); 
	si.cb = sizeof(si); 

	TCHAR* commandLine = GetCommandLine();
	r = CreateProcess(
		NULL,           // Application name
		commandLine,    // Command line
		NULL,           // Process attributes
		NULL,           // Thread attributes
		FALSE,          // Inherit handles
		CREATE_NEW_CONSOLE, // Creation flags
		NULL,           // Envionment
		NULL,           // Current directory
		&si,            // Startup information
		&pi);           // Process information
	if (r == 0) {
		fwprintf( stderr, L"CreateProcess failed with %d.\n", GetLastError());
		return 4;
	}

	CloseHandle(pi.hThread);
	watchpeer(pi.hProcess);

	// Connect to the new client.
	BOOL fConnected = ConnectNamedPipe(hPipeM, NULL) ? TRUE : (GetLastError() == ERROR_PIPE_CONNECTED); 
	if (fConnected == 0)
	{
		fwprintf( stderr, L"ConnectNamedPipe failed with %d.\n", GetLastError()); 
		return 5;
	}

	fConnected = ConnectNamedPipe(hPipeS, NULL) ? TRUE : (GetLastError() == ERROR_PIPE_CONNECTED); 
	if (fConnected == 0)
	{
		fwprintf( stderr, L"ConnectNamedPipe failed with %d.\n", GetLastError()); 
		return 6;
	}

	return dualpump(hPipeM, hPipeS);
}


int connectToPipeServer()
{
	size_t requiredSize;
	TCHAR sPid[20];

	// Get the value of the environment variable.
	_wgetenv_s( &requiredSize, sPid, 20, TEXT(ENVNAME) );

	if( sPid == NULL )
	{
		fwprintf( stderr, L"GetEnv failed with %d.\n", GetLastError()); 
		return 7;
	}

	int pid = _wtoi(sPid);
	if( pid == 0 )
	{
		fwprintf( stderr, L"wtoi failed with %d.\n", GetLastError()); 
		return 8;
	}

	HANDLE hPeer = OpenProcess(SYNCHRONIZE, FALSE, pid);
	watchpeer(hPeer);

	TCHAR sPipeM[PIPENAMESIZE];
	TCHAR sPipeS[PIPENAMESIZE];

	swprintf_s(sPipeM, PIPENAMESIZE, L"\\\\.\\pipe\\concat-%s-M", sPid);
	swprintf_s(sPipeS, PIPENAMESIZE, L"\\\\.\\pipe\\concat-%s-S", sPid);

	HANDLE hPipeM = CreateFile( 
		sPipeM,         // pipe name 
		GENERIC_WRITE,  // write access 
		0,              // no sharing 
		NULL,           // default security attributes
		OPEN_EXISTING,  // opens existing pipe 
		0,              // default attributes 
		NULL);          // no template file 

	// Break if the pipe handle is valid. 
	if (hPipeM == INVALID_HANDLE_VALUE) 
	{
		fwprintf( stderr, L"CreateFile failed with %d.\n", GetLastError()); 
		return 9;
	}

	HANDLE hPipeS = CreateFile( 
		sPipeS,         // pipe name 
		GENERIC_READ,   // read access 
		0,              // no sharing 
		NULL,           // default security attributes
		OPEN_EXISTING,  // opens existing pipe 
		0,              // default attributes 
		NULL);          // no template file 

	// Break if the pipe handle is valid. 
	if (hPipeS == INVALID_HANDLE_VALUE) 
	{
		fwprintf( stderr, L"CreateFile failed with %d.\n", GetLastError()); 
		return 10;
	}

	return dualpump(hPipeS, hPipeM);
}

int dualpump(HANDLE input, HANDLE output)
{
	DWORD dwThreadId;

	HANDLE hThread = CreateThread( 
		NULL,              // no security attribute 
		0,                 // default stack size 
		threadpump,        // thread proc
		(LPVOID)output,    // thread parameter 
		0,                 // not suspended 
		&dwThreadId);      // returns thread ID 

	if (hThread == NULL) 
	{
		printf("CreateThread failed");
		return 11;
	}

	HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	int r = singlepump(input, hStdOut);
	CloseHandle(input);

	WaitForSingleObject(hThread, INFINITE);
	CloseHandle(hThread);

	return r;
}

DWORD WINAPI threadpump(LPVOID lpvParam)
{
	HANDLE output = (HANDLE)lpvParam;
	HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);
	int r = singlepump(hStdIn, output);
	CloseHandle(output);
	return r;
}

int singlepump(HANDLE input, HANDLE output)
{
	DWORD bytesRead;
	DWORD bytesWritten;
	TCHAR buffer[BUFSIZE];
	int r;

	for (;;)
	{
		r = ReadFile(input, &buffer, sizeof(buffer), &bytesRead, NULL);

		if (r == 0)
		{
			r = GetLastError();
			if (r == 109) return 0;
			fwprintf( stderr, L"ReadFile failed with %d.\n", r);
			return 12;
		}

		if (bytesRead == 0) break;

		r = WriteFile(output, &buffer, bytesRead, &bytesWritten, NULL);
		FlushFileBuffers(output);

		if (r == 0)
		{
			fwprintf( stderr, L"WriteFile failed with %d.\n", GetLastError());
			return 13;
		}
	}
	return 14;
}

void watchpeer(HANDLE hProcess)
{
	DWORD dwThreadId;

	HANDLE hThread = CreateThread( 
		NULL,              // no security attribute 
		0,                 // default stack size 
		threadwatchpeer,   // thread proc
		(LPVOID)hProcess,  // thread parameter 
		0,                 // not suspended 
		&dwThreadId);      // returns thread ID 

	if (hThread == NULL) 
	{
		printf("CreateThread failed with %d.\n", GetLastError());
		return;
	}

	CloseHandle(hThread);
}

DWORD WINAPI threadwatchpeer(LPVOID lpvParam)
{
	HANDLE hProcess = (HANDLE)lpvParam;

	WaitForSingleObject(hProcess, INFINITE);

	DWORD exitCode;
	GetExitCodeProcess(hProcess, &exitCode);

	ExitProcess(exitCode);

	return 15;
}